# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# ======================================================================

__version__= '3.5'

from Globals import HTML, HTMLFile, DTMLFile, MessageDialog, Persistent
from ImageFile import ImageFile  
import OFS,Acquisition, AccessControl.Role
import Globals, sys, traceback
from OFS.DTMLDocument import DTMLDocument
from AccessControl import getSecurityManager


from OFS.Document import Document

import os,time,string
from Products.ZSQLMethods.SQL import SQL

from OFS.History import Historical, html_diff
from OFS.SimpleItem import SimpleItem
from OFS.Cache import Cacheable












class SQLFormsDocument(OFS.SimpleItem.Item,DTMLDocument):
    """SQL Forms object."""    
       
    meta_type = 'SQL Form'
    icon = 'misc_/SQLForms/icon'     

 
    manage_about = HTMLFile('README', globals())
    manage_editfields = HTMLFile('dtml/edit_fields', globals())
    manage_methods = HTMLFile('dtml/methods', globals())
    manage_opt = HTMLFile('dtml/options', globals())
    datasel = HTMLFile('dtml/datasel', globals())
    errorinternalmethod = HTMLFile('dtml/error', globals())
    
    def index_html(self):
        """ default html file returned when view object """
        return "nothing here"


    #manage tabs
    manage_options=(
                    {'label':'Options','action':'manage_opt','help':('SQLForms','options.stx')},
                    {'label':'Fields', 'action':'manage_editfields','help':('SQLForms','fields.stx')},
                    {'label':'Methods', 'action':'manage_methods','help':('SQLForms','methods.stx')},
                    #{'label':'Description', 'action':'manage_main'},
                    #{'label':'Properties', 'action':'manage_propertiesForm'},
                    #{'label':'View', 'action':''},
                    #{'label':'Proxy', 'action':'manage_proxyForm', 'help': ('OFSP','DTML-DocumentOrMethod_Proxy.stx')},
                    
                   ) + Historical.manage_options + SimpleItem.manage_options + Cacheable.manage_options
                   


    #security permissions   
#    __ac_permissions__=( 
#        ('Change SQLForms Document',   ('manage_edit', 'manage_upload', 'PUT')),
#        ('Manage SQLForms Document', ('manage_tabs', 'manage_main','manage_about','manage_elementHelp')),
#        )



    sql_types=['varchar','char','integer','float','-ignored-']
    form_types=['text','password','date','textarea','select','list']
    opt_types={'st':'Static text','inc':'Incremental number','dict':'Python dictionary',
               'sqt':'SQL table query','cu':'Current user','cd':'Current date',
               'sc':'Script (dtml or py)','ign':'Field Ignored'}
    unique_opts={'none':'Normal','pri':'Primary key','uniq':'Unique','notnull':'Not null'}
    date_format_dict={
                      1:['%d/%m/%Y',' - dd/mm/yyyy'],
                      2:['%m/%d/%Y',' - mm/dd/yyyy'],
                      3:['%Y/%m/%d',' - yyyy/mm/dd'],
                      4:['%Y/%d/%m',' - yyyy/dd/mm'],
		      5:['%Y/%m/%d-%H:%M:%S',""],
                      10:['%A, %B %d, %Y',""],
                      11:['%a, %b %d, %Y',""],
                      12:['%d %B, %Y',""],
                      13:['%d %b, %Y',""],
                      }
    sql_server_type_dict={
        1:'Generic SQL Server',
        2:'Microsoft SQL Server',
        3:'Oracle SQL Server',
        4:'Postgres SQL Server',
        5:'MySQL SQL Server'
        }

    def __init__(self, id, title,database,sq_table):
        """ init method """
        self.id=id
        self.globals=''
        self._vars=''
        self.version=3
        self.__fields=[
            {
             #field 1
             'fieldname':'autoid','fieldlabel':'key',
             'sq_type':'integer','sq_size':1,
             'form_type':'text','form_size':1,
             'options':'inc','expr':'100',
             'unique':'pri','mandatory':0},
             #field 2
            {
             'fieldname':'uname','fieldlabel':'Your name',
             'sq_type':'varchar','sq_size':30,
             'form_type':'text','form_size':40,
             'options':'st','expr':'Type your name here',
             'unique':'none','mandatory':1},
            ]
        self.primary_key='autoid'
        #create default values
        self.manage_opt_action(title,database,sq_table)
        self.raw="""
Type your comments about this form here.
Type description of this form here.

Here is the body of this document.
"""

    def need_update(self):
        """ check if current sql form instance need conversion """
        flag=0
        if not hasattr(self,'version'): flag=1
        elif int(self.version)<3: flag=1
        return flag

    def manage_update_action(self,REQUEST=None):
        """ Convert old coded forms
        to new style preserving data
        and adding new (and missing) properties """
        
        for i in self.__fields:
            if not hasattr(self,'version'):
                #what isn't until ver 3.0
                self.date_format = '%Y/%m/%d'
                self.table_cell_spacing = 3
                self.textarea_rows = 3
                self.internal_title_color = "#eeeeee"
                self.internal_c1_color = "#ffeeee"
                self.internal_c2_color = "#eeeeff"
                
                i['expr']=i['options']
                i['options']='st'
                if i['sq_type'] == 'incremental':
                    i['options'] = 'inc'
                    i['sq_type'] = 'integer'
                if i['sq_type'] == 'date':
                    i['form_type']='date'
                    i['sq_size']='1'
                    i['sq_type']='integer'
                if i['sq_type'] == 'currentdate':
                    i['options']='cd'
                    i['sq_size']='1'
                    i['sq_type']='integer'
                if i['sq_type'] == 'username':
                    i['options']='cu'
                    i['sq_size']='30'
                    i['sq_type']='varchar'
                if i['form_type'] == 'result':
                    i['form_type']='text'
                if i['form_type'] in ['select','list']: i['options'] = 'sqt'
                if self.primary_key == i['fieldname']: i['unique'] = 'pri'
                else: i['unique']='none'
                
        self.version=3
        self._create_methods()
        self._create_SQL_methods()
        
        if REQUEST is not None:
            return Globals.MessageDialog( title='Updated', message='<strong>%s</strong> has been updated.' % self.id, action ='manage_editfields',)
        


    def manage_opt_action(self,title='',database='',sq_table='no-table',security_add='',security_list='',security_edit='',denied_method='',hide_key='yes',headers='yes',viewlistlen=10,internal_title_color='#eeeeee',internal_c1_color='#ffeeee',internal_c2_color='#eeeeff',date_format='%Y/%m/%d',table_cell_spacing=3,textarea_rows=3,sql_server_type=1,REQUEST=None):
        """ update properties  """
        self.title=title
        
        self.sq_table=sq_table
        self.database=database
        self.security_add=security_add
        self.security_list=security_list
        self.security_edit=security_edit
        self.denied_method=denied_method
        self.hide_key=hide_key
        self.headers=headers
        self.viewlistlen=viewlistlen
        self.internal_title_color=internal_title_color
        self.internal_c1_color=internal_c1_color
        self.internal_c2_color=internal_c2_color
        #added starting with ver 3
        self.date_format=date_format
        self.table_cell_spacing=table_cell_spacing
        self.textarea_rows=textarea_rows
        self.sql_server_type=sql_server_type
        
        self._create_methods()
        self._create_SQL_methods()
        if REQUEST is not None:
            return Globals.MessageDialog( title='Edited', message='<strong>%s</strong> has been edited.' % self.id, action ='manage_opt',)




    

    def fields(self):
        return self.__fields

    def set_fields(self,L):
	self.__fields=L
    


    def manage_order_fields(self,up='',down='',REQUEST=None):
        """ Change order of fields """
        L=self.__fields
        if up!='':
            up=int(up)
            if up>0:
                tmp=L[up]
                del L[up]
                L.insert(up-1,tmp)
                del tmp
        if down!='':
            down=int(down)
            if down<(len(L)-1):
                tmp=L[down]
                del L[down]
                L.insert(down+1,tmp)
                del tmp
        self.__fields=L
        del L
        self._create_methods()
	if REQUEST is not None:
	    return self.manage_editfields(self,REQUEST)





    def manage_edit_fields(self,act='',fieldname=[],fieldlabel=[],sq_type=[],sq_size=[],form_type=[],form_size=[],options=[],expr=[],unique=[],mandatory=[],ix=[],REQUEST=None):
        """ Edit/Update field list """
        is_primary_key=0
        if act == 'Add/Update':
            while '' in fieldname:  fieldname.remove('')
            D=[]
            for i in range(len(fieldname)):
                if i in mandatory: mandatory_val = 1
                else:              mandatory_val = 0
                try:
                    sqsize=int(sq_size[i])
                    if sqsize < 1 : sqsize=1
                except:
                    sqsize = 1

                try:    formsize=int(form_size[i])
                except: formsize = 1

                sqtype = sq_type[i]
                val_expr=expr[i]
                if fieldlabel[i] == '': field_label = fieldname[i]
                else: field_label = fieldlabel[i]

                #
                # all SQL table query   must return key value as integer
                # all Python dictionary must return key value as integer
                # because
                # all 'select' are stored as integer
                if form_type[i] == 'select':
                    sqtype = 'integer'

                #all lists are stored as string of integers separated by a comma
                if form_type[i] == 'list':
                    sqtype = 'varchar'
                    if sqsize < 30 : sqsize = 30 #allocate space

                #make enough space for date stored as string
                if sqtype in ['char','varchar']:
                    if form_type[i] == 'date' or options[i] == 'cd':
                        if sqsize < 20 : sqsize = 20

                #make enough space for current user and store as string
                if options[i] == 'cu':
                    if sqsize < 20 : sqsize = 20
                    sqtype = 'varchar'

                #auto incremental numbers are integers
                if options[i] == 'inc':
                    sqtype = 'integer'
                    try:
                        default_value=int(val_expr)
                    except:
                        default_value=100
                    val_expr=default_value
                
                #static text
                #if options[i] == 'st':
                #    if sqtype not in ['char','varchar']: sqtype = 'varchar'

                #numbers has size = 1
                if sqtype in ['integer','float']: sqsize = 1

                #find unique, primary, ...
                unique_val = unique[i]
                if unique[i]=='pri':
                    if is_primary_key==0:
                        is_primary_key=1
                        self.primary_key=fieldname[i]
                    else:
                        unique_val='none'

                
                if val_expr=='':
                    #create default for 'extra options' field (expr)
                    #opt_types={'st':'Static text','inc':'Incremental number','dict':'Python dictionary','sqt':'SQL table query','cu':'Current user','cd':'Current date'}
                    #if options[i]=='st': val_expr='static text'
                    if options[i]=='cu': val_expr='current user'
                    if options[i]=='sqt': val_expr='sql_table,key,value,where_condition,sort_options'
                    if options[i]=='dict': val_expr="{1:'one',5:'five',2:'two'}"
                    if options[i]=='inc': val_expr='100'
                    if options[i]=='cd': val_expr='current date'
                    if options[i]=='ign': val_expr='this field is not updated by this wizard; is ignored.'
                    if options[i]=='sc': val_expr='place a script-call here'

                D.append({'fieldname':fieldname[i],'fieldlabel':field_label,'sq_type':sqtype,'sq_size':sqsize,'form_type':form_type[i],'form_size':formsize,'options':options[i],'expr':val_expr,'unique':unique_val,'mandatory':mandatory_val})
            self.__fields=D
            del D

        if act == 'Delete':
            if ix is not None:
               D=self.__fields
               ix.reverse()
               for i in ix: del D[i]
            self.__fields=D
            del D
            

        self._create_methods()
        self._create_SQL_methods()
	if REQUEST is not None:
	    return self.manage_editfields(self,REQUEST)





    def _create_methods(self):
        """ create dtml methods """



        ##########################################################
        #                                                        #
        #                                                        #
        #         CREATE list                                    #
        #                                                        #
        #                                                        #
        ##########################################################
        body=''
        if(self.headers=='yes'): body=body+"""<dtml-var standard_html_header>
<dtml-if %s>
""" % (self.security_list,)
        body=body+"""
<dtml-unless REFERER><dtml-call "REQUEST.set('REFERER',HTTP_REFERER)"></dtml-unless>
<dtml-unless sqf_edit_method><dtml-call "REQUEST.set('sqf_edit_method','edit')"></dtml-unless>
"""
        if(self.headers=='yes'): body=body+"""
<h1><dtml-var title_or_id> (List)</h1>
<a href="&dtml-URL2;/">Up</a> -
<a href="&dtml-REFERER;">Back</a> -
<a href="list?REFERER=&dtml-REFERER;">List</a> -
<a href="add?REFERER=&dtml-REFERER;">Add</a>

<br><br>
"""
        body=body+"""
        
<table CELLSPACING='&dtml-table_cell_spacing;'>
<tr BGCOLOR='&dtml-internal_title_color;'>"""
        for i in self.__fields:
                if i['sq_type'] == '-ignored-': continue
                if self.hide_key=='yes' and i['fieldname']==self.primary_key:
                        #hide key
                        continue
                if int(i['form_size'])==0:
                    #skip when hidden field
                    continue
                if int(i['form_size'])<0:
                    #start auth code
                    body=body+"""\n<dtml-if "REQUEST.AUTHENTICATED_USER.has_role('Manager')">"""
                body=body+"""\n <th>%s</th>""" % (i['fieldlabel'])
                if int(i['form_size'])<0:
                    #end auth code
                    body=body+"""\n</dtml-if>"""
        body=body+"""
</tr>


<dtml-call "REQUEST.set('bat_size',_.int(viewlistlen))">
<dtml-unless start_key><dtml-call "REQUEST.set('start_key',0)"></dtml-unless>
<dtml-if "_.int(start_key)<0"><dtml-call "REQUEST.set('start_key',0)"></dtml-if>
<dtml-call "REQUEST.set('next_key',_.int(start_key))">

<dtml-call "REQUEST.set('flag',0)">

<dtml-call "REQUEST.set('q','select * from %s where %s >= '+_.str(start_key)+' order by %s asc')">


<dtml-try>
<dtml-in "SQL(q=q)" size=bat_size>
<dtml-if "_.int(next_key) < _.int(%s)">
  <dtml-call "REQUEST.set('next_key',%s)">
</dtml-if>

<tr valign="top" align="left" BGCOLOR='<dtml-if "_.int(flag)==1">&dtml-internal_c1_color;<dtml-else>&dtml-internal_c2_color;</dtml-if>'>
""" % (self.sq_table,self.primary_key,self.primary_key,
       self.primary_key,self.primary_key)
        for i in self.__fields:
            end_str=""
            if i['sq_type'] == '-ignored-': continue
            if self.hide_key=='yes' and i['fieldname'] == self.primary_key:
                    #hide key
                    continue
            if int(i['form_size']) == 0:
                #skip when hidden field
                continue
            
            if int(i['form_size']) < 0:
                #start auth code
                body = body + """\n<dtml-if "REQUEST.AUTHENTICATED_USER.has_role('Manager')">"""
                end_str = "\n</dtml-if>"
                
            if i['fieldname'] == self.primary_key:
                #provide edit link!!!!
                body = body + """\n <td><a href="&dtml-sqf_edit_method;?edit_key=<dtml-var %s>&REFERER=&dtml-REFERER;"><dtml-var %s></a></td>""" % (i['fieldname'],i['fieldname'])
                body = body + end_str
                continue
                
            if i['form_type'] == 'date' or i['options'] == 'cd':
                if i['sq_type'] in ['integer','float']:
                    #is possibile to be a date field stored as number?
                    body = body + """\n <td><dtml-var "_.DateTime(_.int(%s)).strftime('%s')"></td>""" % (i['fieldname'],self.date_format)
                else:
                    #date field is stored as string
                    body=body+"""\n <td><dtml-var %s></td>""" % (i['fieldname'])
                body = body + end_str

                    
            elif i['form_type'] == 'text':
                body=body+"""\n <td><dtml-var %s></td>""" % (i['fieldname'])

            elif i['form_type'] == 'password':
                body=body+"""\n <td>******</td>"""

            elif i['form_type'] == 'textarea':
                body=body+"""\n <td><dtml-var %s newline_to_br></td>""" % (i['fieldname'])

            elif i['form_type'] == 'select':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts=string.split(i['expr']+',,,,',',')
                    table=opts[0]
                    f_code=opts[1]
                    f_name=opts[2]
                    body=body + """
 <td>
 <dtml-try>
 <dtml-in "SQL(q='select %s as SQLFORM_NAME from %s where %s = '+_.str(%s))">
   <dtml-var SQLFORM_NAME>
 </dtml-in>
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>
</td>""" % (f_name,table,f_code,i['fieldname'])

                else:
                    #assume there is a python dict!
                    body = body + '''
 <td>
<dtml-try>
   <dtml-call "REQUEST.set('%s_expr',  %s  )">
   <dtml-var "%s_expr[_.int(%s)]">
<dtml-except>
  <dtml-var errorinternalmethod>
</dtml-try>
 </td>''' % (i['fieldname'],i['expr'],i['fieldname'],i['fieldname'])



              
            elif i['form_type'] == 'list':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts = string.split(i['expr']+',,,,',',')
                    table = opts[0]
                    f_code = opts[1]
                    f_name = opts[2]
                    body = body+"""
 <td>
 <dtml-try>
 <dtml-if "%s!='0'">
 <ul>
 <dtml-in "_.string.split(%s,',')">
   <dtml-in "SQL(q='select %s as SQLFORM_NAME from %s where %s = '+_.str(_['sequence-item']))">
    <li><dtml-var SQLFORM_NAME></li>
   </dtml-in>
 </dtml-in>
 </ul>
 </dtml-if>
 <dtml-except>
    <dtml-var errorinternalmethod>
 </dtml-try>
 </td>
""" % (i['fieldname'],i['fieldname'],f_name,table,f_code)
                else:
                    #assume there is a python dict!
                    body = body + """
 <td>
 <dtml-try>
   <dtml-if "%s!='0'">
   <dtml-call "REQUEST.set('%s_expr',  %s  )">
   <dtml-call "REQUEST.set('%s_list',_.string.split(%s,','))">
     <ul>
     <dtml-in %s_list>
       <li><dtml-var "%s_expr[_.int(_['sequence-item'])]"></li>
     </dtml-in>
     </ul>
    </dtml-if>
 <dtml-except>
    <dtml-var errorinternalmethod>
 </dtml-try>
 </td>
""" % (
       i['fieldname'],
       i['fieldname'],i['expr'],
       i['fieldname'],i['fieldname'],
       i['fieldname'],i['fieldname'],)
                    
            
            body = body + end_str


        body=body+"""
</tr>
<dtml-call "REQUEST.set('flag',1-_.int(flag))">
</dtml-in>

<dtml-except>
  <h1>General error</h1>
  <dtml-var errorinternalmethod>
</dtml-try>

</table>
<!-- navigation links -->
<br>
<a href="&dtml-URL0;?start_key=0&REFERER=&dtml-REFERER;">First</a> -
<a href="&dtml-URL0;?start_key=<dtml-var "_.int(start_key)-_.int(bat_size)">&REFERER=&dtml-REFERER;">Prev</a> -
<a href="&dtml-URL0;?start_key=<dtml-var "_.int(next_key)">&REFERER=&dtml-REFERER;">Next</a>
"""
        if(self.headers=='yes'): body=body+"""
<dtml-else>
    <dtml-var %s>
</dtml-if>
<dtml-var standard_html_footer>
""" % (self.denied_method,)
        self.list = Document(body, __name__='list')





        ##########################################################
        #                                                        #
        #                                                        #
        #         CREATE add                                     #
        #                                                        #
        #                                                        #
        ##########################################################
        body=''
        if(self.headers=='yes'): body=body+"""<dtml-var standard_html_header>
<dtml-if %s>
""" % (self.security_add,)
        body=body+"""
<dtml-unless REFERER><dtml-call "REQUEST.set('REFERER',HTTP_REFERER)"></dtml-unless>
<dtml-if add_data>
<dtml-unless submit><dtml-call "REQUEST.set('submit','Add')"></dtml-unless>
<dtml-if "submit!='Cancel'">
"""
        sql_server_type = int(self.sql_server_type)
        
        for i in self.__fields:
            #skip ignored fields
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            
            #current date
            if i['options'] == 'cd':
                if i['sq_type'] in ['char','varchar']: expression="_.str(_.DateTime().strftime('%s'))" % (self.date_format)
                else: expression="_.int(_.DateTime())"
                body = body + """<dtml-call "REQUEST.set('%s',%s)">
""" % (i['fieldname'],expression)

            #current user
            if i['options'] == 'cu':
                body = body + """<dtml-call "REQUEST.set('%s',REQUEST.AUTHENTICATED_USER.getUserName())">
""" % (i['fieldname'])

            #generate next value (is sql_max()+1)
            if i['options'] == 'inc' and sql_server_type == 1: #only for generic SQL server
                body = body + """
<dtml-try>
  <dtml-in "SQL(q='select max(%s) as sqlfmaxid from %s')">
     <dtml-unless sqlfmaxid><dtml-call "REQUEST.set('sqlfmaxid','%s')"></dtml-unless>
     <dtml-with REQUEST>
         <dtml-call "REQUEST.set('sqlf_internal_newid',_.int(sqlfmaxid)+1)">
     </dtml-with>
  <dtml-else>
     <dtml-call "REQUEST.set('sqlf_internal_newid','%s')">
  </dtml-in>
<dtml-except>
   <dtml-call "REQUEST.set('sqlf_internal_newid','%s')">
</dtml-try>

<dtml-call "REQUEST.set('%s',REQUEST['sqlf_internal_newid'])">
""" % (i['fieldname'],self.sq_table,i['expr'],(int(i['expr'])+1),(int(i['expr'])+1),i['fieldname'],)

            #scripts (dtml or py)
            if i['options'] == 'sc':
                body = body + """<dtml-call "REQUEST.set('%s',%s)">
""" % (i['fieldname'],i['expr'])
            
                
            
            #hidden fields (with form_size=0)
            #noobody can edit them.
            #(used maybe for compatibility with older applications which require that field in SQL table)
            #will get default value from 'expr' internal field
            if int(i['form_size'])==0:
                body = body + """<dtml-call "REQUEST.set('%s','%s')">
""" % (i['fieldname'],i['expr'])


            #when field is date - store it as string or integer
            if i['form_type'] == 'date':
                if i['sq_type'] in ['char','varchar']:
                    #str
                    expression="_.str(_.DateTime(_.str(%s)).strftime('%s'))" % (i['fieldname'],self.date_format)
                else:
                    #int
                    expression="_.int(_.DateTime(_.str(%s)))" % (i['fieldname'])
                body = body + """<dtml-call "REQUEST.set('%s',%s)">
""" % (i['fieldname'],expression)

            
            
            if i['form_type'] == 'list':
                #if is a list and nothing checked -> it should be empty list
                body = body + """<dtml-unless %s><dtml-call "REQUEST.set('%s',['0'])"></dtml-unless>
""" % (i['fieldname'],i['fieldname'])
            else:
                #generate default value (just in case)
                #but avoid generation of default for auto filled fields or pre processed fields
                if i['options'] not in ['inc','cu','cd'] and i['form_type'] not in ['date']:
                    if i['sq_type'] in ['char','varchar']:
                        default=''
                        body = body + """<dtml-unless %s><dtml-call "REQUEST.set('%s','%s')"></dtml-unless>
""" % (i['fieldname'],i['fieldname'],default)
                    else:
                        default = 0
                        body = body + """<dtml-unless %s><dtml-call "REQUEST.set('%s',%s)"></dtml-unless>
""" % (i['fieldname'],i['fieldname'],default)

            #if is a number - what if it has wrong format?
            if i['sq_type'] in ['integer','float'] and i['options'] != 'inc':
                if i['options'] == 'st' : default=i['expr']
                else: default="0"
                body = body + """<dtml-try><dtml-call "REQUEST.set('%s',_.int(%s))"><dtml-except><dtml-call "REQUEST.set('%s',%s)"></dtml-try>
""" % (i['fieldname'],i['fieldname'],i['fieldname'],default)
                


                
        body=body+"""
<dtml-call "REQUEST.set('q','insert into %s ( """ % (self.sq_table)
        for i in self.__fields:
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            if i['options'] == 'inc':
                if sql_server_type !=3 and sql_server_type != 1: continue #only for != generic SQL server and Oracle
            body=body+"%s," % (i['fieldname'])
            
        body=body[:-1]+") values('+"
        
        for i in self.__fields:
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            if i['options'] == 'inc':
                if sql_server_type!=3 and sql_server_type != 1: continue #only for != generic SQL server and Oracle
            if i['options'] == 'inc':
                #for Oracle only
                if sql_server_type == 3:
                    body = body + "'%s_seq_%s.nextval,'+" % (self.sq_table,i['fieldname'])
                    continue
            if i['sq_type'] in ['integer','float']:
                body=body+"_.str(REQUEST['%s'])+','+" % (i['fieldname'])
            elif i['form_type']=='list':
                body=body+"'\\''+_.string.join(REQUEST['%s'],',')+'\\','+" % (i['fieldname'])
            else:
                #body=body+"'\\''+_.string.replace(_.str(REQUEST['%s']),'\\'','\\'\\'')+'\\','+" % (i['fieldname'])
                body=body+"'\\''+_.string.replace(_.str(REQUEST['%s']),'\\'','\\'\\'')+'\\','+" % (i['fieldname'])
        body=body[:-3]+""")')">
        
 
<dtml-try>
  <dtml-call "SQL(q=q)">
  <dtml-if "act=='Add&Continue'">
    <dtml-call "RESPONSE.redirect(URL0+'?REFERER='+REFERER)">
  <dtml-else>
    <dtml-call "RESPONSE.redirect(REFERER)">
  </dtml-if>
  <dtml-comment>
  <dtml-var q>
  </dtml-comment>

  
  
<dtml-except>
  <dtml-var errorinternalmethod>
  SQL query used :<br>
  <dtml-var q newline_to_br><br>
  <br>
  Go <a href="JavaScript:history.go(-1)">back</a> and try again!
</dtml-try>

</dtml-if>



<dtml-else>
<script>
function parse_date(s)
 {
 i1=s.indexOf("/");
 i2=i1+(s.slice(i1+1)).indexOf("/")+1;
 if (i1<0 || i2<0)
  {
   year=<dtml-var "ZopeTime().strftime('%Y')">;
   month=<dtml-var "ZopeTime().strftime('%m')">-1;
   day=<dtml-var "ZopeTime().strftime('%d')">;
  }
 else
  {
   year=s.slice(0,i1);
   month=s.slice(i1+1,i2)-1;
   day=s.slice(i2+1);
  }
 return 'datasel?year='+year+'&month='+month+'&day='+day;
 }
</script>
"""
        if(self.headers=='yes'): body=body+"""
<h1><dtml-var title_or_id> (Add)</h1>
<a href="&dtml-URL2;/">Up</a> -
<a href="&dtml-REFERER;">Back</a> -
<a href="list?REFERER=&dtml-REFERER;">List</a> -
<a href="add?REFERER=&dtml-REFERER;">Add</a>
<br>
"""
        #
        #  build input form
        #
        body=body+"""
<dtml-call "REQUEST.set('flag',0)">
<form method="post" name="sqlformsadd">
<input type="hidden" name="REFERER" value="<dtml-var REFERER>">
<input type="hidden" name="add_data" value="true">
<table CELLSPACING='&dtml-table_cell_spacing;'>
<tr>
 <th> </th>
 <td>
 <input type="button" value="Add&Continue" onclick="check_data('Add&Continue')">
 &nbsp;&nbsp;&nbsp;&nbsp;
 <input type="button" value="Add" onclick="check_data('Add')">
 &nbsp;&nbsp;&nbsp;&nbsp;
 <input type="button" value="Cancel" onclick="location.href='&dtml-REFERER;';">
 </td>
</tr>

"""
        for i in self.__fields:
            #skip 'current date', 'current user', 'incremental number', 'script (dtml or py)' and 'Ignored' fields
            #build value in action method of form
            if i['options'] in ['cd','cu','inc','sc','ign']: continue
            
            #skip hidden fields (with form_size=0)
            #build value in action method of form
            if int(i['form_size'])==0: continue
            
            #negative size means that only admin can see and change field
            if int(i['form_size'])<0:
                form_size=-int(i['form_size'])
                body=body+"""
<dtml-if "REQUEST.AUTHENTICATED_USER.has_role('Manager')">"""
            else: form_size=int(i['form_size'])

            body=body+'''
<tr valign="top" align="left" BGCOLOR='<dtml-if "_.int(flag)==1">&dtml-internal_c1_color;<dtml-else>&dtml-internal_c2_color;</dtml-if>'>
<dtml-call "REQUEST.set('flag',1-_.int(flag))">
 <th align="left">%s</th>
 <td>
 <dtml-try>
 ''' % (str(i['fieldlabel']))
            s='(unknown form type)'
            if i['form_type'] == 'text':
                s='<input type="text" name="%s" size="%s" maxlength="%s" value="<dtml-if %s><dtml-var %s><dtml-else>%s</dtml-if>">' % (i['fieldname'],str(form_size),str(form_size),i['fieldname'],i['fieldname'],str(i['expr']),)
            if i['form_type'] == 'password':
                s='<input type="password" name="%s" size="%s" maxlength="%s" value="<dtml-if %s><dtml-var %s><dtml-else>%s</dtml-if>">' % (i['fieldname'],str(form_size),str(form_size),i['fieldname'],i['fieldname'],str(i['expr']),)
            if i['form_type'] == 'date':
                s='''
 <input type="text" name="view_%s" size="%s" value="<dtml-if %s><dtml-var %s><dtml-else>%s</dtml-if>" disabled>
 <input type="hidden" name="%s" value="<dtml-if %s><dtml-var %s><dtml-else>%s</dtml-if>">
 
 <input type=button value="..." onClick='window.retfield = document.forms[0].view_%s; window.datafield = document.forms[0].%s; window.open(parse_date(window.datafield.value), "chooser", "toolbar=no,menubar=no,scrollbars=yes,width=300,height=300")'>
''' % (i['fieldname'],str(form_size),i['fieldname'],i['fieldname'],str(i['expr']),
            i['fieldname'],i['fieldname'],i['fieldname'],str(i['expr']),
            i['fieldname'],i['fieldname'],)
                
            if i['form_type'] == 'textarea':
                s='<textarea name="%s" cols="%s" rows="&dtml-textarea_rows;"><dtml-if %s><dtml-var %s><dtml-else>%s</dtml-if></textarea>' % (i['fieldname'],str(form_size),i['fieldname'],i['fieldname'],str(i['expr']),)

            if i['form_type'] == 'select':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts=string.split(i['expr']+',,,,',',')
                    table=opts[0]
                    f_code=opts[1]
                    f_name=opts[2]
                    f_where=opts[3]
                    f_sort=opts[4]
                    if f_where!='':
                        f_where='where '+f_where
                    if f_sort!='':
                        f_sort='order by '+f_sort
                    s='''
  <select name="%s">
   <dtml-in "SQL(q='select %s as SQLFORM_CODE,%s as SQLFORM_NAME from %s %s %s')">
     <option value="<dtml-var SQLFORM_CODE>"<dtml-if %s><dtml-if "_.str(%s)==_.str(SQLFORM_CODE)"> SELECTED</dtml-if></dtml-if>><dtml-var SQLFORM_NAME></option>
   </dtml-in>
  </select>
''' % (i['fieldname'],f_code,f_name,table,f_where,f_sort,i['fieldname'],i['fieldname'])
                else:
                    #assume there is a python dict!
                    s='''
  <dtml-call "REQUEST.set('%s_expr',%s)">
  <dtml-call "REQUEST.set('%s_keys',%s_expr.keys())">
  <dtml-call "%s_keys.sort()">
  <select name="%s">
   <dtml-in "%s_keys">
     <option value="<dtml-var sequence-item>"<dtml-if %s><dtml-if "_.str(%s)==_.str(_['sequence-item'])"> SELECTED</dtml-if></dtml-if>><dtml-var "%s_expr[_['sequence-item']]"></option>
   </dtml-in>
  </select>
''' % (i['fieldname'],i['expr'],       #REQ set 1
             i['fieldname'],i['fieldname'],  #REQ set 2
             i['fieldname'], #sort
             i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'])



            if i['form_type'] == 'list':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts=string.split(i['expr']+',,,,',',')
                    table=opts[0]
                    f_code=opts[1]
                    f_name=opts[2]
                    f_where=opts[3]
                    f_sort=opts[4]
                    if f_where!='':
                        f_where='where '+f_where
                    if f_sort!='':
                        f_sort='order by '+f_sort
                    s='''
   <dtml-in "SQL(q='select %s as SQLFORM_CODE,%s as SQLFORM_NAME from %s %s %s')">
     <input type="checkbox" name="%s:list:string" value="<dtml-var SQLFORM_CODE>" <dtml-if %s><dtml-if "_.str(SQLFORM_CODE)==_.str(%s) or _.str(SQLFORM_CODE) in _.string.split(_.str(%s),',')"> CHECKED</dtml-if></dtml-if>><dtml-var SQLFORM_NAME><br>
   </dtml-in>
''' % (f_code,f_name,table,f_where,f_sort,i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'])
                else:
                    #assume there is a python dict!
                    s='''
   <dtml-call "REQUEST.set('%s_expr',%s)">
   <dtml-call "REQUEST.set('%s_keys',%s_expr.keys())">
   <dtml-call "%s_keys.sort()">
   <dtml-in "%s_keys">
     <input type="checkbox" name="%s:list:string" value="<dtml-var sequence-item>"<dtml-if %s><dtml-if "_.str(%s)==_.str(_['sequence-item']) or _.str(_['sequence-item']) in _.string.split(_.str(%s),',')"> CHECKED</dtml-if></dtml-if>><dtml-var "%s_expr[_['sequence-item']]"><br>
   </dtml-in>
 ''' % (i['fieldname'],i['expr'],       #REQ set 1
             i['fieldname'],i['fieldname'],  #REQ set 2
             i['fieldname'], #sort
             i['fieldname'], #dtml-in
             i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'])
 
            
            
            body=body+s+'''
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>
 </td>
</tr>
'''
            if int(i['form_size'])<0:
                body=body+"""<dtml-else>
  <input type="hidden" name="%s" value="%s">
</dtml-if>""" % (i['fieldname'],i['expr'],)
        body=body+"""

<tr>
 <th> </th>
 <td>
 <input type="hidden" name="act" value="0">
 <input type="button" value="Add&Continue" onclick="check_data('Add&Continue')">
 &nbsp;&nbsp;&nbsp;&nbsp;
 <input type="button" value="Add" onclick="check_data('Add')">
 &nbsp;&nbsp;&nbsp;&nbsp;
 <input type="button" value="Cancel" onclick="location.href='&dtml-REFERER;';">
 </td>
 
</tr>
</table>
</form>

<script>
function check_data(s)
{
"""
        #build javascript check routine
        for i in self.__fields:
            if int(i['mandatory'])==0: continue #jump to next field
            if i['form_type'] in ['select','list']: continue
            if i['options'] in ['inc','cd','cu']: continue
            body=body+"""
  if (document.sqlformsadd.%s.value=='%s')
     {
        alert('A value for\\n%s\\nis mandatory!');
        document.sqlformsadd.%s.focus();
        return;
     } """ % (i['fieldname'],i['expr'],i['fieldlabel'],i['fieldname'],)
        body=body+"""
  document.forms.sqlformsadd.act.value=s;
  document.forms.sqlformsadd.submit();
}
</script>
</dtml-if>
"""
        if(self.headers=='yes'): body=body+"""
<dtml-else>
    <dtml-var %s>
</dtml-if>
<dtml-var standard_html_footer>
""" % (self.denied_method,)
        self.add = Document(body, __name__='add')








        ##########################################################
        #                                                        #
        #                                                        #
        #         CREATE edit                                    #
        #                                                        #
        #                                                        #
        ##########################################################
        
        body=''
        if(self.headers=='yes'): body=body+"""<dtml-var standard_html_header>
<dtml-if %s>
""" % (self.security_edit)
        body=body+"""
<dtml-unless REFERER><dtml-call "REQUEST.set('REFERER',HTTP_REFERER)"></dtml-unless>
<dtml-if submit>
<dtml-if "submit=='Delete'">"""
        body=body+"""
<dtml-call "REQUEST.set('q','DELETE from %s where %s = '+_.str(edit_key))">
<dtml-try>
  <dtml-call "SQL(q=q)">
<dtml-except>
  <dtml-var errorinternalmethod>
  SQL query used :<br>
  <dtml-var q newline_to_br><br>
  <br>
  Go <a href="JavaScript:history.go(-1)">back</a> and try again!
</dtml-try>
</dtml-if delete>

<dtml-if "submit=='Edit'">
""" % (self.sq_table,self.primary_key)


        for i in self.__fields:
            #skip ignored fields
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            
            #skip 'current date', 'current user' and 'incremental number' fields
            if i['options'] in ['cd','cu','inc']: continue

            #skip hidden fields (with form_size=0)
            if int(i['form_size'])==0: continue

            #when field is date - store it as string or integer
            if i['form_type'] == 'date':
                if i['sq_type'] in ['char','varchar']:
                    #str
                    expression="_.str(_.DateTime(_.str(%s)).strftime('%s'))" % (i['fieldname'],self.date_format)
                else:
                    #int
                    expression="_.int(_.DateTime(_.str(%s)))" % (i['fieldname'])
                body = body + """<dtml-call "REQUEST.set('%s',%s)">
""" % (i['fieldname'],expression)

            #if is a list and nothing checked -> it should be empty list (lists should be stored as SQL string)
            if i['form_type'] == 'list':
                body = body + """<dtml-unless %s><dtml-call "REQUEST.set('%s',['0'])"></dtml-unless>
""" % (i['fieldname'],i['fieldname'])

            #if is a number - what if it has wrong format?
            if i['sq_type'] in ['integer','float']:
                if i['options'] == 'st' : default=i['expr']
                else: default="0"
                body = body + """<dtml-try><dtml-call "REQUEST.set('%s',_.int(%s))"><dtml-except><dtml-call "REQUEST.set('%s',%s)"></dtml-try>
""" % (i['fieldname'],i['fieldname'],i['fieldname'],default)

            #scripts (dtml or py)
            if i['options'] == 'sc':
                body = body + """<dtml-call "REQUEST.set('%s',%s)">
""" % (i['fieldname'],i['expr'])
                


            
    
        #for admins
        body=body+"""
    <dtml-if "REQUEST.AUTHENTICATED_USER.has_role('Manager')">
    <dtml-call "REQUEST.set('q','UPDATE %s SET """ % (self.sq_table)
        for i in self.__fields:
            #skip ignored field
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            
            #skip 'current date', 'current user' and 'incremental number' fields
            if i['options'] in ['cd','cu','inc']: continue

            #skip hidden fields (with form_size=0)
            if int(i['form_size'])==0: continue
            
            if i['form_type']=='list':
                value="_.string.join(REQUEST['%s'],',')" % (i['fieldname'],)
            else:
                value="_.str(REQUEST[\'"+i['fieldname']+"\'])"
                
            if i['sq_type'] in ['char','varchar']:
                body=body+"""%s=\\''+_.string.replace(%s,'\\'','\\'\\'')+
'\\',""" % (i['fieldname'],value)
            else:
                body=body+"""%s='+%s+
',""" % (i['fieldname'],value)
        
        body = body[:-1]+""" where %s = '+_.str(edit_key))">

    <dtml-else>
""" % (self.primary_key)
        #for non admins
        body = body + """<dtml-call "REQUEST.set('q','UPDATE %s SET """ % (self.sq_table)
        for i in self.__fields:
            #skip ignored field
            if i['options'] == 'ign': continue
            if i['sq_type'] == '-ignored-': continue
            
            #skip 'current date', 'current user' and 'incremental number' fields
            if i['options'] in ['cd','cu','inc']: continue

            #skip hidden fields (with form_size=0) and slso with form_size<0
            if int(i['form_size'])<=0: continue
            
            if i['form_type']=='list':
                value="_.string.join(REQUEST['%s'],',')" % (i['fieldname'],)
            else:
                value="_.str(REQUEST[\'"+i['fieldname']+"\'])"
                
            if i['sq_type'] in ['char','varchar']:
                body=body+"""%s=\\''+_.string.replace(%s,'\\'','\\'\\'')+
'\\',""" % (i['fieldname'],value)
            else:
                body=body+"""%s='+%s+
',""" % (i['fieldname'],value)
        
        body = body[:-1]+""" where %s = '+_.str(edit_key))">
    </dtml-if>
<dtml-try>
  <dtml-call "SQL(q=q)">
<dtml-except>
  <dtml-var errorinternalmethod>
  SQL query used :<br>
  <dtml-var q newline_to_br><br>
  <br>
  Go <a href="JavaScript:history.go(-1)">back</a> and try again!
</dtml-try>
</dtml-if edit>

<dtml-call "RESPONSE.redirect(REFERER)">


""" % (self.primary_key)
        body = body + """

<dtml-else>
<script>
function parse_date(s)
 {
 i1=s.indexOf("/");
 i2=i1+(s.slice(i1+1)).indexOf("/")+1;
 if (i1<0 || i2<0)
  {
   year=<dtml-var "ZopeTime().strftime('%Y')">;
   month=<dtml-var "ZopeTime().strftime('%m')">-1;
   day=<dtml-var "ZopeTime().strftime('%d')">;
  }
 else
  {
   year=s.slice(0,i1);
   month=s.slice(i1+1,i2)-1;
   day=s.slice(i2+1);
  }
 return 'datasel?year='+year+'&month='+month+'&day='+day;
 }
</script>
"""

        if(self.headers=='yes'): body=body+"""
<h1><dtml-var title_or_id> (Edit)</h1>
<a href="&dtml-URL2;/">Up</a> -
<a href="&dtml-REFERER;">Back</a> -
<a href="list?REFERER=&dtml-REFERER;">List</a> -
<a href="add?REFERER=&dtml-REFERER;">Add</a>
<br>"""

        #
        #build html form
        #
        body=body+"""
<dtml-call "REQUEST.set('flag',0)">
<form method="post">
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Edit">
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Delete">
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Cancel">
<br>
<table CELLSPACING='&dtml-table_cell_spacing;'>
<input type="hidden" name="REFERER" value="<dtml-var REFERER>">
<input type="hidden" name="edit_key" value="<dtml-var edit_key>">
<dtml-in "SQL(q='select * from %s where %s = '+_.str(edit_key))">""" % (self.sq_table,self.primary_key)
        for i in self.__fields:
            #skip ignored field
            #skip 'current date', 'current user', 'incremental number', 'script (dtml or py)' and 'Ignored' fields
            if i['options'] in ['cd','cu','inc','sc','ign']: continue
            
            #generate default values for ignored fields
            if i['sq_type'] == '-ignored-':
                body = body + """<dtml-var "REQUEST.set('%s','%s')">\n""" % (i['fieldname'],i['expr'])
            
            #skip hidden fields (with form_size=0)
            if int(i['form_size'])==0: continue            
            
            #negative size means that only admin can see and change field
            if int(i['form_size'])<0:
                  form_size=-int(i['form_size'])
                  body=body+"""\n<dtml-if "REQUEST.AUTHENTICATED_USER.has_role('Manager')">"""  
            else: form_size=int(i['form_size'])
            
            body=body+"""
  <tr BGCOLOR='<dtml-if "_.int(flag)==1">&dtml-internal_c1_color;<dtml-else>&dtml-internal_c2_color;</dtml-if>'>
    <dtml-call "REQUEST.set('flag',1-_.int(flag))">
    <th align="left" valign="top">%s</th>
    <td align="left" valign="top">""" % (i['fieldlabel'])
            s='unknown type'
            if i['form_type'] == 'text':
                s='<input type="text" name="%s" size="%s" maxlength="%s" value="&dtml-%s;">' % (i['fieldname'],str(form_size),str(form_size),i['fieldname'],)
            if i['form_type'] == 'password':
                s='PPP<input type="password" name="%s" size="%s" maxlength="%s" value="&dtml-%s;">' % (i['fieldname'],str(form_size),str(form_size),i['fieldname'],)
            if i['form_type'] == 'date':
                if i['sq_type'] in ['char','varchar']:
                    s='''
 <input type="text" name="view_%s" size="%s" value="<dtml-var %s>" disabled>
 <input type="hidden" name="%s" value="<dtml-var %s>">
 <input type="button" onClick='window.retfield = document.forms[0].%s; window.datafield = document.forms[0].view_%s; window.open(parse_date(window.datafield.value), "chooser", "toolbar=no,menubar=no,scrollbars=yes,width=300,height=300")' value="...">
''' % (i['fieldname'],str(form_size),str(i['fieldname']),
            i['fieldname'],str(i['fieldname']),
            i['fieldname'],
            i['fieldname'],)
                else:
                    s='''
 <input type="text" name="view_%s" size="%s" value="<dtml-var "ZopeTime(%s).strftime('%%Y/%%m/%%d')">" disabled>
 <input type="hidden" name="%s" value="<dtml-var "ZopeTime(%s).strftime('%%Y/%%m/%%d')">">
 <input type="button" onClick='window.retfield = document.forms[0].%s; window.datafield = document.forms[0].view_%s; window.open(parse_date(window.datafield.value), "chooser", "toolbar=no,menubar=no,scrollbars=yes,width=300,height=300")' value="...">
''' % (i['fieldname'],str(form_size),str(i['fieldname']),
            i['fieldname'],str(i['fieldname']),
            i['fieldname'],
            i['fieldname'],)

            if i['form_type'] == 'textarea':
                s='<textarea name="%s" cols="%s" rows="&dtml-textarea_rows;">&dtml-%s;</textarea>' % (i['fieldname'],str(form_size),i['fieldname'],)
                
            if i['form_type'] == 'select':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts=string.split(i['expr']+',,,,',',')
                    table=opts[0]
                    f_code=opts[1]
                    f_name=opts[2]
                    f_where=opts[3]
                    f_sort=opts[4]
                    if f_where!='':
                        f_where='where '+f_where
                    if f_sort!='':
                        f_sort='order by '+f_sort
                    s='''
 <dtml-try>
  <select name="%s">
   <dtml-in "SQL(q='select %s as SQLFORM_CODE,%s as SQLFORM_NAME from %s %s %s')">
     <option value="<dtml-var SQLFORM_CODE>"<dtml-if %s><dtml-if "_.str(%s)==_.str(SQLFORM_CODE)"> SELECTED</dtml-if></dtml-if>><dtml-var SQLFORM_NAME></option>
   </dtml-in>
  </select>
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>
''' % (i['fieldname'],#select
       f_code,f_name,table,f_where,f_sort,#SQL in
       i['fieldname'],i['fieldname'])
                else:
                    #assume there is a python dict!
                    s='''
 <dtml-try>
  <dtml-call "REQUEST.set('%s_expr',%s)">
  <dtml-call "REQUEST.set('%s_keys',%s_expr.keys())">
  <dtml-call "%s_keys.sort()">
  <select name="%s">
   <dtml-in "%s_keys">
     <option value="<dtml-var sequence-item>"<dtml-if %s><dtml-if "_.str(%s)==_.str(_['sequence-item'])"> SELECTED</dtml-if></dtml-if>><dtml-var "%s_expr[_['sequence-item']]"></option>
   </dtml-in>
  </select>
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>
 </td>''' % (i['fieldname'],i['expr'],       #REQ set 1
             i['fieldname'],i['fieldname'],  #REQ set 2
             i['fieldname'], #sort
             i['fieldname'], #select
             i['fieldname'], #dtml-in
             i['fieldname'],i['fieldname'],i['fieldname'])


            if i['form_type'] == 'list':
                if i['options'] == 'sqt':
                    #assume there is an SQL table query
                    opts=string.split(i['expr']+',,,,' , ',')
                    table=opts[0]
                    f_code=opts[1]
                    f_name=opts[2]
                    f_where=opts[3]
                    f_sort=opts[4]
                    if f_where!='':
                        f_where='where '+f_where
                    if f_sort!='':
                        f_sort='order by '+f_sort
                    s='''
 <dtml-try>
   <dtml-in "SQL(q='select %s as SQLFORM_CODE,%s as SQLFORM_NAME from %s %s %s')">
     <input type='checkbox' value="<dtml-var SQLFORM_CODE>" name="%s:list:string"<dtml-if %s><dtml-if "_.str(SQLFORM_CODE) in _.string.split(_.str(%s),',')"> CHECKED</dtml-if></dtml-if>><dtml-var SQLFORM_NAME><br>
   </dtml-in>
   <br>
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>

''' % (f_code,f_name,table,f_where,f_sort,#SQL in
       i['fieldname'],i['fieldname'],i['fieldname'])
                else:
                    #assume there is a python dict!
                    s='''
 <dtml-try>
   <dtml-call "REQUEST.set('%s_expr',%s)">
   <dtml-call "REQUEST.set('%s_keys',%s_expr.keys())">
   <dtml-call "%s_keys.sort()">
   <dtml-in "%s_keys">
     <input type="checkbox" name="%s:list:string" value="<dtml-var sequence-item>"<dtml-if %s><dtml-if "_.str(%s)==_.str(_['sequence-item']) or _.str(_['sequence-item']) in _.string.split(_.str(%s),',')"> CHECKED</dtml-if></dtml-if>><dtml-var "%s_expr[_['sequence-item']]"><br>
   </dtml-in>
 <dtml-except>
  <dtml-var errorinternalmethod>
 </dtml-try>
 </td>''' % (i['fieldname'],i['expr'],       #REQ set 1
             i['fieldname'],i['fieldname'],  #REQ set 2
             i['fieldname'], #sort
             i['fieldname'], #dtml-in
             i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'],i['fieldname'])


 
            body=body+s+"</td>\n  </tr>"
            if int(i['form_size'])<0:
                body=body+"\n</dtml-if>"
        body=body+"""
</table>
<br>
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Edit">
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Delete">
&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
<input type="submit" name="submit" value="Cancel">
</form>
"""
        
        body=body+"""
</dtml-in>
</dtml-if>
"""
        if(self.headers=='yes'): body=body+"""
<dtml-else>
   <dtml-var %s>
</dtml-if>
<dtml-var standard_html_footer>
""" % (self.denied_method,)
        self.edit = Document(body, __name__='edit')








    def _create_SQL_methods(self):
        """ create SQL internal methods """
        #Generic SQL method
        self.SQL=SQL('SQL', 'Generic SQL Method', self.database, 'q', '''
<dtml-comment>
  This is a generic SQL method.
  Is used to be called from other methods.
  DO NOT EDIT THIS METHOD!
</dtml-comment>

<dtml-var q>
''')
        sql_server_type = int(self.sql_server_type)
        #SQL create sequence
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        for i in self.__fields:
            if i['options'] == 'inc':
                try: start_val = int(i['expr'])
                except: start_val = 1
                if sql_server_type == 3: #Oracle
                    q = q + 'create sequence %s_seq_%s start with %d\n' % (self.sq_table,i['fieldname'],start_val)
                if sql_server_type == 4: #Postgres
                    q = q + 'create sequence %s_seq_%s start %d\n' % (self.sq_table,i['fieldname'],start_val)
        self.SQLCreateSEQ=SQL('SQLCreateSEQ', 'Create sequence', self.database, '', q)
        
        #SQL create table
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
                
        q = q + '''
create table '''+str(self.sq_table)+'\n('
        for i in self.__fields:
                if i['sq_type'] == '-ignored-': continue
                if i['sq_type'] == 'float':
                    sq_type='float'
                    sq_size=1
                elif i['sq_type'] == 'integer':
                    sq_type='integer'
                    sq_size=1
                else:
                    sq_type=i['sq_type']
                    if i['sq_size']=='': sq_size=0
                    else: sq_size=int(i['sq_size'])
                q=q+'\n  '+str(i['fieldname'])+' '+str(sq_type)
                if sq_size>1: q=q+'('+str(sq_size)+')'
                if i['options'] == 'inc':
                    try: start_val = int(i['expr'])
                    except: start_val = 1
                    #if sql_server_type == 1 -> Generic SQL Server                        
                    if sql_server_type == 2: #Microsoft SQL
                        q=q+' IDENTITY (%d, 1)' % (start_val)
                    #if sql_server_type == 3: #Oracle - this (as default) should work, but it don't! So I'll use seq.nextval in insert statement
                    #    q=q+' default %s_seq_%s.nextval' % (self.sq_table,i['fieldname'])
                    if sql_server_type == 4: #Postgres
                        q=q+" default nextval('%s_seq_%s')" % (self.sq_table,i['fieldname'])
                    if sql_server_type == 5: #MySQL
                        q=q+' auto_increment default %d' % (start_val)
                    
                        
                if i['unique']=='uniq': q=q+' UNIQUE'
                if i['unique']=='pri': q=q+' PRIMARY KEY'
                if i['unique']=='notnull': q=q+' NOT NULL'
                #if i['unique']=='none': q=q+' NULL'
                q=q+','
        q=q[:-1]+'\n)'
        self.SQLCreate=SQL('SQLCreate', 'Create Table', self.database, '', q)
        
        #SQL drop sequence
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        for i in self.__fields:
            if i['options'] == 'inc':
                if sql_server_type == 3: #Oracle
                    q = q + 'drop sequence %s_seq_%s\n' % (self.sq_table,i['fieldname'])
                if sql_server_type == 4: #Postgres
                    q = q + 'drop sequence %s_seq_%s\n' % (self.sq_table,i['fieldname'])
        self.SQLDropSEQ=SQL('SQLDropSEQ', 'Drop sequence', self.database, '', q)
        
        #SQL drop table
        q='''
<dtml-comment>
  This is an internal method.
  DO NOT EDIT THIS METHOD!
  Do NOT call it directelly!
  SQL Server: %s
</dtml-comment>

''' % (self.sql_server_type_dict[sql_server_type])
        q = q + '''drop table '''+str(self.sq_table)
        self.SQLDrop=SQL('SQLDrop', 'Drop Table', self.database, '', q)
        


        

    def manage_sql(self,act='',REQUEST=None):
        """ create/drop table """
        self._create_SQL_methods()
        sql_server_type = int(self.sql_server_type)
        if act == 'Create table':
            try:
                if sql_server_type in [3,4]: self.SQLCreateSEQ() #only for postgres and oracle
                self.SQLCreate()
                return Globals.MessageDialog( title='Create!', message="<p align='left'>Ok! Table created.<br><br>See generated SQL method <a href='SQLCreate/manage_main'>here</a>!</p>", action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Create', message="<p align='left'><font color='red'>Error! Can't create table.<br>See generated SQL method <a href='SQLCreate/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)
            
        if act == 'Drop table':
            try:
                if sql_server_type in [3,4]: self.SQLDropSEQ() #only for postgres and oracle
                self.SQLDrop()
                return Globals.MessageDialog( title='Drop', message="<p align='left'>Ok! Table erased (dropped).<br><br>See generated SQL method <a href='SQLDrop/manage_main'>here</a>!</p>", action ='manage_opt',)
            except:
                type, val, tb = sys.exc_info()
                errmsg=(string.join(traceback.format_exception(type, val, tb), ''))
                return Globals.MessageDialog( title='Drop', message="<p align='left'><font color='red'>Error! Table may not exist.<br>Or somebody locked it.<br><br>See generated SQL method <a href='SQLDrop/manage_main'>here</a>!<br><br><div align='left'><pre>"+errmsg+"</pre></div></font></p>", action ='manage_opt',)


        
        return Globals.MessageDialog( title='Oops', message='Oops! Unexpected argument!', action ='manage_opt',)
            

    






Globals.default__class_init__(SQLFormsDocument)



def ReturnIDs(self,lista):
      """Find Scripts in the current folder and above

    This function return a list of ids.
    """
      ids={}
      have_id=ids.has_key
      StringType=type('')
      while self is not None:
          if hasattr(self, 'objectValues'):
              #lista=['DTML Method','Script (Python)','External Method']
              for o in self.objectValues(lista):
                      id=o.id
                      if type(id) is not StringType: id=id()
                      if not have_id(id):
                          if hasattr(o,'title'):
                              title=o.title
                              if type(title) is not StringType: title=title()
                              if title == '': title=id
                          else: title=id
                          ids[id]=[id,title]
          if hasattr(self, 'aq_parent'): self=self.aq_parent
          else: self=None

      ids=map(lambda item: (item[1], item[0]), ids.items())
      #ids.sort()
      return ids

